;;;; "simsynch.scm" simulate synchronous systems.	-*-scheme-*-
;;; Copyright (C) 1995, 1996, 1997 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(display ";			 loading SIMSYNCH ")
(display
 (call-with-input-file (in-vicinity (program-vicinity) "version.txi")
   (lambda (port) (read port) (read port) (read port))))
(newline)
(require 'database-browse)
(define b browse)

;;(provide 'check-turnaround)

(require 'sort)
(require 'line-i/o)
(require 'database-utilities)
(require 'common-list-functions)
(require 'logic)			;(in-vicinity (program-vicinity) "logic.scm")
(require 'models)			;(in-vicinity (program-vicinity) "models.scm")

(define (create-board board-name dbtype)
  (define boardb (create-database (string-append
				   (symbol->string board-name) ".db")
				  dbtype))

  ((boardb 'add-domain)
   '(optstring #f (lambda (x) (or (not x) (string? x))) string #f))
  (define-tables boardb
    '(constants
      ((name symbol))
      ((value expression)
       (optcom optstring))
      ())

    '(ptag
      ((name symbol))
      ((pld-type string)
       (signature expression (lambda (x) (or (not x) (number? x))))
       (optcom optstring))
      ((int "" #f #f)
       (hid "" #f #f)
       (* "*" #f #f)))

    '(pin-type
      ((name atom))
      ()
      ((output)
       (input/output)
       (wire-and)
       (wire-or)
       (clock)
       (input)
       (unused)
       (internal)))

    '(pin
      ((ptag ptag)
       (name symbol))
      ((pin-type pin-type)
       (pin-name expression)		;'?, number, symbol, or list of same.
       (output-enable expression)
       (paired-input expression (lambda (x) (or (boolean? x) (symbol? x)))))
      ())

    '(pindex
      ((ptag ptag)
       (pin-name atom))
      ((symbol-name symbol)
       (pin-type pin-type)
       (name expression)		;corresponds to NAME in pin table.
       (signal-name expression))
      ())

    '(pin-basename
      ((ptag ptag)
       (basename symbol))		;name without bounds
      ((names expression (lambda (x) (every symbol? x))))
      ())

    '(block
      ((name symbol))
      ((clock expression)
       (reset symbol)			;really signal foreign-key
       (clock-frequency expression)
       (synchronous-reset? boolean)
       (optcom optstring))
      ())

    '(check
      ((block block)
       (order uint))
      ((expression expression))
      ())

    '(signal-type
      ((name symbol))
      ()
      ((registered)
       (registered-input)
       (combinatorial)
       (hidden)				;used for inputs
       (macro)))

    '(signal
      ((block block)
       (name symbol))
      ((defining-ptag ptag)
       (signal-type signal-type)
       (initial-state expression)
       (next-function expression)
       (order uint)
       (optcom optstring))
      ())

    '(used-in
      ((name symbol)
       (block block))
      ()
      ())

    '(flags
      ((block block)
       (signal-name symbol))
      ((flags expression))
      ())

    '(u-ptag
      ((u-number symbol))
      ((ptag ptag))
      ())

    '(netlist
      ((name symbol)
       (index atom))
      ((sheet atom)
       (u-pins expression
	       (lambda (x)
		 (and (list? x)
		      (every
		       (lambda (elt)
			 (and (pair? elt) (symbol? (car elt))
			      (or (integer? (cdr elt))
				  (symbol? (cdr elt)))))
		       x))))
       (r-pins expression
	       (lambda (x) (and (list? x) (every integer? x)))))
      ())

    '(netlist-aggregate
      ((name symbol))
      ((bounds expression
	       (lambda (b) (if b (and (list? b)
				      (integer? (car b))
				      (integer? (cadr b)))
			       #t))))
      ())

    '(netlist-non-aggregate
      ((name symbol))
      ()
      ())

    '(netlist-inverse
      ((ptag ptag)
       (pin atom (lambda (x) (or (integer? x) (symbol? x)))))
      ((net-name symbol)
       (index atom))
      ())

    '(resistor-nets
      ((number uint))
      ((names expression
	      (lambda (x) (and (list? x) (every symbol? x)))))
      ())
    )
  boardb)

(define *block* #f)
(defvar *design* 'test)
(define *board* (create-board *design* 'alist-table))

(define (synch:register-block block . args)
  (let* ((clock (if (null? args)
		    (symbol-append block '- 'clock)
		    (car args)))
	 (reset (if (or (null? args) (null? (cdr args)))
		    (symbol-append block '- 'reset-)
		    (cadr args)))
	 (synchronous-reset?
	  (and (pair? reset)
	       (eq? (car reset) 'SYNCHRONOUS)
	       (begin (set! reset (cadr reset))
		      #t)))
	 (clock-frequency #f))
    ((((*board* 'open-table) 'block #t) 'row:insert)
     (list block clock reset clock-frequency synchronous-reset? (comment)))
    (set! *block* block)))

(define (synch:register-ptag tag ptag . args)
  (define signature (if (null? args) #f (car args)))
  ((((*board* 'open-table) 'ptag #t) 'row:insert)
   (list tag ptag signature (comment))))

(define ptag-table ((*board* 'open-table) 'ptag #t))
(define ptag->pld (ptag-table 'get 'pld-type))
(define ptag->sig (ptag-table 'get 'signature))

(define pin-table ((*board* 'open-table) 'pin #t))
(define pin-table:row-insert (pin-table 'row:insert))
(define pin-table:row-retrieve (pin-table 'row:retrieve))

(define pin-basename-table ((*board* 'open-table) 'pin-basename #t))
(define pin-basename:row-retrieve (pin-basename-table 'row:retrieve))
(define pin-basename:update
  (let ((pin-basename:row-update (pin-basename-table 'row:update)))
    (lambda (ptag name pin-type)
      (let* ((bnds (vector-bounds name))
	     (basename (if bnds (car bnds) name))
	     (row (pin-basename:row-retrieve
		   ptag (if bnds (car bnds) name))))
	(if row
	    (or (eq? pin-type 
		     ((pin-table 'get 'pin-type) ptag (caaddr row)))
		(synch:warn "Pins with same basename but different types"
			    ptag name (caaddr row))))
	(pin-basename:row-update 
	 (list ptag basename 
	       (if row (cons name (caddr row)) (list name))))))))

(define ordering:counter 100)
(define signal-table ((*board* 'open-table) 'signal #t))
(define signal-table:row-insert (signal-table 'row:insert))
(define signal-table:row-remove (signal-table 'row:remove))
(define signal-table:row-retrieve (signal-table 'row:retrieve))
(define signal-table:get-init-fun (signal-table 'get 'initial-state))
(define signal-table:get-next-fun (signal-table 'get 'next-function))

(define flags-table ((*board* 'open-table) 'flags #t))
(define flags:row-update (flags-table 'row:update))
(define flags:row-retrieve (flags-table 'row:retrieve))
(define (synch:signal-flags tag name)
  (let ((row (flags:row-retrieve tag name)))
    (if row (list-ref row 2) '())))

(define used-in-table ((*board* 'open-table) 'used-in #t))
(define used-in-table:row-insert (used-in-table 'row:insert))
;;(define used-in-table:row-remove (used-in-table 'row:remove))
(define used-in-table:row-retrieve* (used-in-table 'row:retrieve*))
(define used-in-table:get-block (used-in-table 'get 'block))

(define block-table ((*board* 'open-table) 'block #f))

(define (mark-uses block next-function)
  (cond ((symbol? next-function)
	 (if (not (used-in-table:get-block next-function block))
	     (used-in-table:row-insert (list next-function block))))
	((and (list? next-function) (pair? next-function))
	 (for-each (lambda (expr) (mark-uses block expr))
		   next-function))))

(define (used-by-other-blocks? signal block)
  (not (null? (used-in-table:row-retrieve*
	       signal (lambda (key) (not (eq? key block)))))))

(define constant-table ((*board* 'open-table) 'constants #t))
(define constant-table:insert
  (let ((inserter (constant-table 'row:insert)))
    (lambda args
      (inserter args))))
(define constant-table:row-retrieve (constant-table 'row:retrieve))

(defmacro defconst (name value)
  (cond ((list? name) `(defconst ,(car name) (lambda ,(cdr name) ,value)))
	(else (cond ((not (slib:eval `(defined? ,name))))
		    ((and (symbol? name) (eqv? (slib:eval value)
					       (slib:eval name))))
		    (else (slib:error 'trying-to-defconst name
				      'to-different-value value)))
	      `(begin (define ,name ,value)
		      (constant-table:insert ',name ,name ,(comment))))))
(defconst *xxxx* #x-f)

;;; Eval constant
(define (ceval thing)
  (cond ((number? thing) thing)
	((not (symbol? thing)) thing)
	((constant-table:row-retrieve thing) =>
	 cadr)
	(else thing)))

(define (for-each-ptag proc)
  ((ptag-table 'for-each-row)
   (lambda (ptag-row)
     (define str (cadr ptag-row))
     (if (and (not (equal? "" str))
	      (or (char-alphabetic? (string-ref str 0))
		  (char-numeric? (string-ref str 0))))
	 (apply proc ptag-row)))))

(define check-table ((*board* 'open-table) 'check #t))
(define synch:insert-check
  (let ((check-table:row-insert (check-table 'row:insert)))
    (lambda (check)
      (set! ordering:counter (+ ordering:counter 1))
      (check-table:row-insert (list *block* ordering:counter check)))))

;;(SYNCH:DEFINE-PIN
;; ptag name pin-type pin-name output-enable paired-input)
;;(SYNCH:DEFINE-SIGNAL
;; [block] name defining-ptag signal-type initial-state next-function)
(define (synch:define-pin ptag . args)
  (cond (ptag
	 (if (not ((ptag-table 'row:retrieve) ptag))
	     (synch:register-ptag ptag ""))
	 (pin-table:row-insert (cons ptag args))
	 (apply
	  (lambda (name pin-type pin-name output-enable paired-input)
	    (pin-basename:update ptag name pin-type)
	    (cond
	     ((eq? pin-name '?))
	     ((and (or (list? pin-name) (symbol? pin-name))
		   (vector-bounds (or paired-input name)))
	      => (lambda (bounds)
		   (let* ((first (cadr bounds))
			  (last (caddr bounds))
			  (range (- last first))
			  (incer (sign range)))
		     (do ((i 0 (+ incer i)))
			 ((positive? (* incer (- i range))))
		       (pindex:row-insert
			(list ptag
			      (if (pair? pin-name)
				  (list-ref pin-name (abs i))
				  (symbol-append pin-name (abs i)))
			      '?
			      pin-type
			      name
			      (refvectorize (or paired-input name) (+ i first))
			      ))))))
	     ((or (and (number? pin-name) (not (zero? pin-name)) pin-name)
		  (and (symbol? pin-name) (ptag-stripped pin-name)))
	      => (lambda (pname)
		   (pindex:row-insert
		    (list ptag pname '? pin-type name name))))))
	  args))))
(define (synch:define-signal name ptag . args)
  (set! ordering:counter (+ ordering:counter 1))
  (cond (ptag (if (not ((ptag-table 'row:retrieve) ptag))
		  (synch:register-ptag ptag ""))))
  (apply (lambda (block name ptag signal-type initial-state next-function)
	   (cond ((and (list? initial-state)
		       (eq? 'SYNCHRONOUS (car initial-state)))
		  (set! initial-state (cadr initial-state))
		  (synch:add-signal-flag block name 'SYNCHRONOUS-RESET)))
	   (signal-table:row-insert (list block name ptag signal-type
					  initial-state next-function
					  ordering:counter
					  (comment))))
	 *block* name (or ptag '*) args))

(define (synch:define-bus name float-limit . connections)
  (define name-drv (symbol-append name '- 'drv))
  #+check-turnaround
  (define name-lst (symbol-append name '- 'lst))
  (if (eqv? 1 (length connections))
      (set! connections (car connections)))
  (synch:define-signal name 'hid 'macro #f
		       `(cond ,@connections (else ,(- -1 float-limit))))
  #+check-turnaround
  (synch:define-signal
   name-lst 'hid 'registered (- -1 float-limit) name)
  (synch:define-signal
   name-drv 'hid 'registered (- -1 float-limit)
   `(case (count ,@(remove #f (map car connections)))
      ((0) (cond ((eqv? -1 ,name-drv)
		  (synch:warn ',name 'floating ,float-limit 'cycles)
		  -1)
		 ((and (number? ,name-drv) (negative? ,name-drv))
		  (+ 1 ,name-drv))
		 (else ,(- -1 float-limit))))
      ((1) (let ((idx (cond ,@(let ((idx -1))
				(map (lambda (con)
				       (set! idx (+ 1 idx))
				       `(,(car con) ,idx))
				     connections))
			    (else #f))))
	     #+check-turnaround
	     (cond ((eqv? idx ,name-drv))
		   ((negative? ,name-drv))
		   ((negative? idx))
		   ((eqv? ,name ,name-lst))
		   (else (synch:warn ',name 'no-turnaround-time ,name-drv idx)))
	     idx))
      ,@(if (some boolean? (map car connections))
	    `(((2) ,name-drv))
	    '())
      (else (apply synch:warn ',name 'bus-driving-conflict
		   (list ,@(map (lambda (x) `',(car x)) connections)))
	    (cond ,@(let ((idx -1))
		      (map (lambda (con)
			     (set! idx (+ 1 idx))
			     `(,(car con) ,idx))
			   connections))
		  (else #f))))))

(define (synch:add-signal-flag block signal-name flag)
  (let ((row (flags:row-retrieve block signal-name)))
    (flags:row-update (list block signal-name
			    (if row
				(cons flag (list-ref row 2))
				(list flag))))))

(define (synch:check-flags recognized-flags)
  ((flags-table 'for-each-row)
   (lambda (row)
     (apply (lambda (block signal flags)
	      (for-each (lambda (flag)
			  (or (memq flag recognized-flags)
			      (synch:warn "unrecognized flag" flag block signal)))
			flags))
	    row))))

(defmacro synch:defcheck (check)
  (tprint 'synch:defcheck check)
  (synch:insert-check check)
  "check")

;;; Synch:pre is a macro that saves typing.

(defmacro synch:pre (signal)
  (tprint 'synch:pre signal)
  (let ((stepper (signal-table:get-next-fun *block* signal)))
    (case stepper
      ((#f) (synch:error "signal not known:" *block* signal))
      ((?) (synch:error "signal not yet set:" *block* signal))
      (else stepper))))

(defmacro synch:defmacro (name expression)
  (define ptag (ptag-prefix name))
  (set! name (ptag-stripped name))
  (tprint 'synch:defmacro ptag name)
  (synch:define-signal name ptag 'macro #f expression)
  (symbol->string name))

(defmacro synch:defshare (pin-nam name)
  (define ptag (ptag-prefix name))
  (set! name (ptag-stripped name))
  (tprint 'synch:defshare ptag name)
  (cond ((signal-table:row-retrieve *block* name)
	 (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name))
	(else (synch:error "shared signal not defined:" name)))
  (symbol->string name))

(defmacro synch:definput (pin-nam name expression)
  (define ptag (or (ptag-prefix pin-nam) (ptag-prefix name)))
  (if (eq? ptag (ptag-prefix name)) (set! name (ptag-stripped name)))
  (tprint 'synch:definput ptag name)
  (synch:define-signal name ptag 'hidden #f expression)
  (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name)
  (symbol->string name))

(defmacro synch:defreginput (pin-nam fullname expression)
  (let* ((ptag (or (ptag-prefix pin-nam) (ptag-prefix fullname)))
	 (name (ptag-stripped fullname))
	 (fullprename (symbol-append ptag ': 'pre: name)))
    (tprint 'synch:defreginput ptag name)
    `(begin
       (define ,name #f)		;This BAD Hack is spreading!
					;Needed for self-reset test in scm2mach
					;registered inputs clear to #f
       (synch:definput ,(ptag-stripped pin-nam) ,fullprename ,expression)
       (synch:define 0 ,fullname ,name #f #f)
       (synch:set! ,name ,(symbol-append 'pre: name)))))

(defmacro synch:define (pin-nam signal init-state pin oen . opt-next)
  (define ptag (or (ptag-prefix pin-nam) (ptag-prefix signal)))
  (set! opt-next (if (null? opt-next) '? (car opt-next)))
  (tprint 'synch:define ptag signal)
  (cond (ptag
	 (if (eq? ptag (ptag-prefix signal))
	     (set! signal (ptag-stripped signal)))
	 (synch:define-pin ptag signal
			   (case oen
			     ((#f) 'internal)
			     ((#t) 'output)
			     (else 'input/output))
			   (ptag-stripped pin-nam)
			   oen
			   (if oen (or pin signal) pin))))
  (synch:define-signal signal ptag 'registered init-state opt-next)
  (symbol->string signal))

(defmacro synch:set/reset! (signal turn-on turn-off)
  (tprint 'synch:set/reset! signal)
  (synch:insert-check
   `(cond ((and ,turn-on ,turn-off)
	   (synch:warn ',signal 'being-turned-on-and_off_))))
  `(synch:set! ,signal (and (not ,turn-off) (or ,turn-on ,signal))))

(defmacro synch:set! (signal new-state)
  (tprint 'synch:set! signal)
  (let* ((row (signal-table:row-remove *block* signal))
	 (last3 (and row (last row 3))))
    (cond ((not row) (synch:error "signal not known:" signal))
	  ((eq? '? (car last3))
	   (signal-table:row-insert
	    (append (butlast row 3) (cons new-state (cdr last3)))))
	  ((and (pair? (car last3)) (eq? 'compose-sets (caar last3)))
	   (signal-table:row-insert
	    (append (butlast row 3)
		    (cons (append (car last3) `((lambda (,signal) ,new-state)))
			  (cdr last3)))))
	  (else
	   (signal-table:row-insert
	    (append (butlast row 3)
		    (cons `(compose-sets ,signal
					 (lambda (,signal) ,(car last3))
					 (lambda (,signal) ,new-state))
			  (cdr last3))))))
    (symbol->string signal)))

(defmacro synch:declare (signal flag)
  (tprint 'synch:declare signal)
  (synch:add-signal-flag *block* signal flag))

(define synch:step '())
(define synch:status #f)
(defmacro define-synchronous-system (system-name . body)
  (cond ((not (eq? system-name *block*))
	 (slib:error "defining system-name different from *block*?" system-name
		     *block*)))
  (synch:define-signal 'synch:prin-count 'INT 'registered 0 'synch:prin-count)
  (synch:define-signal 'synch:count 'INT 'registered 0 '(+ 1 synch:count))
  (let ((names '())
	(inits '())
	(macros '())
	(hiddens '())
	(steppers '())
	(block->clock (((*board* 'open-table) 'block #f) 'get 'clock))
	(block->reset (((*board* 'open-table) 'block #f) 'get 'reset))
	(synch:blocks ((((*board* 'open-table) 'block #f) 'get* 'name))))
    (for-each
     (lambda (sig-row)
       (apply
	(lambda (block name defining-ptag signal-type
		       initial-state next-function order optcom)
	  (case signal-type
	    ((registered combinatorial macro)
	     (mark-uses block next-function)))
	  (case signal-type
	    ((registered) (if (eq? '? next-function)
			      (synch:error "signal not set:" sig-row))
	     (set! names (cons name names))
	     (set! inits (cons initial-state inits))
	     (set! steppers (cons next-function steppers)))
	    ((hidden) (set! hiddens (cons (list name next-function) hiddens)))
	    ((macro combinatorial)
	     (set! macros (cons (list name next-function) macros)))
	    (else (synch:error "unknown signal-type" sig-row))))
	sig-row))
     (sort ((signal-table 'row:retrieve*) *block*)
	   (lambda (row1 row2) (> (list-ref row1 6) (list-ref row2 6)))))
    ((*board* 'write-database) (string-append (symbol->string *design*) ".db"))
    (set! synch:step
	  (slib:eval
	   `(lambda ()
;;; 	      ,@(map (lambda (sys-name)
;;; 		       (define time-name (symbol-append sys-name ':time))
;;; 		       `(if ,time-name
;;; 			    (set! ,time-name (+ ,time-name synch:inc))))
;;; 		     synch:blocks)
	      (or?
	       ,@(map (lambda (sys-name)
			;;(define time-name (symbol-append sys-name ':time))
			(define clock-expr (block->clock sys-name))
			(define clock-state (symbol-append sys-name ':clock))
			(define reset-name (block->reset sys-name))
			(define state-name (symbol-append sys-name ':state))
			`(cond (,(if (negated? reset-name)
				     `(not ,reset-name)
				     reset-name)
				(set! ,state-name #f)
				#f)
			       ((not ,state-name)
				(set! ,state-name
				      (slib:eval
				       (cons 'list
					     ,(symbol-append
					       sys-name ':state-initializers))))
				;;(if (number? ,clock-expr) (set! ,time-name 0))
				#f)
;;; 			       ((and ,time-name (> ,time-name ,clock-expr))
;;; 				(set! ,time-name (- ,time-name ,clock-expr))
;;; 				(set! ,state-name
;;; 				      (@apply
;;; 				       ,(symbol-append sys-name ':state-step)
;;; 				       ,state-name))
;;; 				#t)
;;; 			       (,time-name #t)
			       ((positive? ,(symbol-append sys-name ':queue))
				(set! ,(symbol-append sys-name ':queue)
				      (+ -1 ,(symbol-append sys-name ':queue)))
				(set! ,state-name
				      (@apply
				       ,(symbol-append sys-name ':state-step)
				       ,state-name))
				#t)
			       (else #f)))
		      synch:blocks)))))	;(reverse )
    (set! synch:status
	  (slib:eval
	   `(lambda ()
	      ;;(display "...")
	      (if synch:stop (printf "STOPPED!"))
	      ,@(map (lambda (block)
		       (let ((reset-name (block->reset block)))
			 `(apply printf " %s: %s q=%d r=%d p=%d"
				 ',(string-upcase (symbol->string block))
				 (if ,(if (negated? reset-name) reset-name
					  `(not ,reset-name))
				     "run"
				     "rst")
				 ,(symbol-append block ':queue)
				 (reverse
				  (last (or ,(symbol-append block ':state)
					    (list 0 0))
					2)))))
		     synch:blocks)	;(reverse )
	      (newline))))
    `(begin
       (define ,(symbol-append system-name ':clock)
	 ',(block->clock system-name))
       ;;(define ,(symbol-append system-name ':time) #f)
       (define ,(symbol-append system-name ':queue) 0)
       (define (,(symbol-append system-name '-sim) num)
	 (set! ,(symbol-append system-name ':queue)
	       (max num ,(symbol-append system-name ':queue)))
	 (kb:announce-polling 'break)
	 (simulate!))
       (define ,(symbol-append system-name ':state) #f)
       (define ,(symbol-append system-name ':state-initializers) ',inits)
       (defmacro ,system-name (name)
	 (if (and (list? name) (pair? name) (eq? 'quote (car name)))
	     (set! name (cadr name)))
	 (let ((vals (symbol-append ',system-name ':state))
	       (idx (position name ',names)))
	   (if (not idx)
	       (slib:error name 'not-registered-in vals)
	       `(if ,vals
		    (list-ref ,vals ,idx)
		    (slib:eval
		     (list-ref
		      ,',(symbol-append system-name ':state-initializers)
		      ,idx))))))
       (define (,(symbol-append system-name ':state-step) ,@names)
	 (fluid-let ((*block* ',*block*))
	   (let* ,macros
	     (let* ,hiddens
	       ,@(map caddr
		      (sort ((check-table 'row:retrieve*) *block*)
			    (lambda (row1 row2) (> (cadr row1) (cadr row2)))))
	       ,@body
	       (let ((new-state (list ,@steppers)))
		 (update-deferred-sets!)
		 new-state))))))))

(define simulating? #f)
(define (simulate!)
  (cond (simulating?)
	(synch:stop (synch:status) (kb:break))
	((kb:break-on-hit))
	(else (set! simulating? #t)
	      (let ((more-to-do? (synch:step)))
		(set! simulating? #f)
		(if more-to-do? (simulate!))))))

(define synch:stop #f)
(define (synch:reset) (set! synch:stop #f))

;;; Print out as many warnings and errors as possible before stopping.
(require 'object->string)
(defvar print-timing #t)
(defvar print-1-timing #f)
(defvar print-chat #t)
(define synch:msg-list '())
(defvar talkative #t)
(define (tprint . args)
  ;;(if talkative (apply print args))
  #f)

(defvar report slib:error)

(define (synch:chat msg)
  (set! synch:msg-list (cons msg synch:msg-list))
  #t)

(define (synch:warn . args)
  (define cep (current-error-port))
  (cond
   ((printf:format? (car args))
    (apply fprintf cep args))
   (else
    (display "WARN: " cep)
    (if (not (null? args))
	(begin (display (car args) cep)
	       (for-each (lambda (x) (display #\  cep) (write x cep))
			 (cdr args))))))
  (newline cep)
  (force-output cep)
  (set! print-1-timing #t))

(define (printf:format? str)
  (cond ((not (string? str)) #f)
	 ((< 100 (string-length str)) #f)
	 (else (memv #\% (string->list str)))))

(define (synch:error . args)
  (define cep (current-error-port))
  (cond
   ((printf:format? (car args))
    (apply fprintf cep args))
   (else
    (display "ERROR: " cep)
    (if (not (null? args))
	(begin (display (car args) cep)
	       (for-each (lambda (x) (display #\  cep) (write x cep))
			 (cdr args))))))
  (newline cep)
  (force-output cep)
  (set! print-1-timing #t)
  (set! synch:stop #t))

(defmacro synch:prin-header (h cnt)
  (cond ((and (list? h) (pair? h) (eq? (car h) 'quote))
	 (set! h (cadr h))))
  (let* ((ss (map (lambda (i)
		    (cond ((string? i) (string-upcase! i))
			  ((char? i) "")
			  (else (object->string i))))
		  h))
	 (sl (map string-length ss))
	 (ml (apply max sl))
	 (ll (map (lambda (s l)
		    (string->list
		     (if (zero? l) (make-string ml #\nul)
			 (string-append s (make-string (- ml l) #\ )))))
		  ss sl))
	 (rl (apply map list ll)))
    `(let ()
       (dprintf "%4d\\n" ,cnt)
       ,@(map
	  (lambda (l) `(dprintf ,l))
	  (map (lambda (l)
		 (string-append
		  (apply
		   string-append
		   " "
		   (if (char=? #\nul (car l)) "" (string (car l)))
		   (map (lambda (c)
			  (string-append
			   "  " (if (char=? #\nul c) "" (string c))))
			(cdr l)))
		  ;;(cond (cnt (display "  ")(display cnt) (set! cnt #f)))
		  "\\n"))
	       rl))
       "synch:prin-header")))

(define (synch:prin1 x)
  (case x
    ((#t) (display "  |"))
    ((#f) (display " | "))
    (else
     (display #\ )
     (cond ((char? x) (display x))
	   ((number? x)
	    (let ((s (number->string x 16))
		  (d 2))
	      (do ((l (string-length s) (+ 1 l)))
		  ((>= l d) (display s))
		(display #\0))))
	   ((string? x) (display x))
	   (else (write x))))))

(define synch:prin-args #f)
(define synch:last-printed-count 0)

(defmacro synch:print args
  (define names '())
  (define argl '())
  (do ((l args (cdr l)))
      ((null? l))
    (cond ((string? (car l))
	   (SET! names (cons (car l) names))
	   (set! argl (cons (cadr l) argl))
	   (set! l (cdr l)))
	  ((char? (car l))
	   (set! names (cons (car l) names))
	   (set! argl (cons (car l) argl)))
	  (else
	   (set! names (cons (object->string (car l)) names))
	   (set! argl (cons (car l) argl)))))
  `(begin
     (cond
      ((or PRINT-TIMING print-1-timing)
       (set! print-1-timing #f)
       (let ((args (list ,@(reverse argl))))
	 (cond ((not (eqv? synch:last-printed-count (+ -1 synch:count)))
		(synch:status)
		))
	 (cond ((or (not (equal? args synch:prin-args))
		    (not (null? synch:msg-list)))	; (and PRINT-CHAT)
		(cond ((zero? (remainder synch:prin-count 50))
		       (synch:prin-header ',(reverse names) synch:count)
		       (newline)))
		(set! synch:prin-count (+ 1 synch:prin-count))
		(set! synch:prin-args args)
		(for-each synch:prin1 args)
		(for-each (lambda (m) (display #\ ) (display m))
			  synch:msg-list)
		(newline))))
       (set! synch:last-printed-count synch:count))
      ((and PRINT-CHAT (or ;;(zero? (remainder synch:count 50))
			   (not (null? synch:msg-list))))
       (printf "%5.5d" synch:count)
       (for-each (lambda (m) (display #\ ) (display m)) synch:msg-list)
       (newline)))
     (set! synch:msg-list '())))

;   ================================================================

(define pindex ((*board* 'open-table) 'pindex #t))
(define pindex:row-insert (pindex 'row:insert))
(define pindex:row-retrieve (pindex 'row:retrieve))

(define (expand-range pair)
  (if (list? pair) pair
      (do ((l '() (cons j l))
	   (inc (if (> (cdr pair) (car pair)) -1 +1))
	   (j (cdr pair) (+ inc j)))
	  ((= j (car pair)) (cons j l)))))

(define (db:read-pins ptag name)
  (define (insert pnum pnam ptyp)
    (let ((s (assq ptyp '((out . output)
			  (i/o . input/output)
			  (hiZ . input/output)
			  ;;(wire-and)
			  ;;(wire-or)
			  (clk . clock)
			  (in . input)
			  (v . unused)
			  ;;(internal)
			  ))))
      (if s
	  (pindex:row-insert (list ptag pnum pnam (cdr s) #f #f))
	  (synch:warn 'db:read-pins "unknown pin type" ptyp))))
  (call-with-input-file (string-append name ".pin")
    (lambda (iport)
      (let ((shape (read iport)))
	(cond
	 ((not (list? shape)) (slib:error "not valid format" shape))
	 ((memq (car shape) '(pqfp plcc lcc soj))
	  (set-car! (cdr shape) (expand-range (cadr shape)))
	  (do ((rows (cadr shape) (cdr rows)))
	      ((null? rows))
	    (let* ((pnum (read iport))
		   (pnam (read iport))
		   (ptyp (read iport)))
	      (cond ((not (equal? pnum (car rows)))
		     (slib:error "pins out of sequence"
				 (car rows) "~=" pnum)))
	      (insert pnum pnam ptyp))))
	 ((memq (car shape) '(bga))
	  (let loop ((c (peek-char iport)))
	    (cond ((eof-object? c))
		  ((char-whitespace? c)
		   (read-char iport)
		   (loop (peek-char iport)))
		  (else
		   (let* ((pnum (read iport))
			  (pnam (read iport))
			  (ptyp (read iport)))
		     (insert pnum pnam ptyp)
		     (loop (peek-char iport)))))))
	 (else (slib:error "unknown format" shape)))))))

(define ptag:row-retrieve (ptag-table 'row:retrieve))
(define u-ptag ((*board* 'open-table) 'u-ptag #t))
(define u-ptag:row-retrieve (u-ptag 'row:retrieve))

(define db:register-u-ptag
  (let ((u-ptag:row-insert (u-ptag 'row:insert)))
    (lambda (u-number ptag)
      (if (ptag:row-retrieve ptag)
	  (u-ptag:row-insert (list u-number ptag))
	  (let* ((chip (object->string ptag))
		 (ptag (symbol-append chip '- u-number)))
	    (synch:register-ptag ptag "")
	    (u-ptag:row-insert (list u-number ptag)))))))

;; Reads a file containing a list of lists (U-NUMBER PTAG-NAME)
(define (db:read-u-ptag name)
  (for-each (lambda (row)
	      (apply db:register-u-ptag row))
	    (call-with-input-file name read)))

(define netlist ((*board* 'open-table) 'netlist #t))
(define netlist:row-retrieve (netlist 'row:retrieve))
(define netlist-inverse ((*board* 'open-table) 'netlist-inverse #t))
(define netlist-inverse:row-retrieve (netlist-inverse 'row:retrieve))
(define netlist-aggregate ((*board* 'open-table) 'netlist-aggregate #t))
(define netlist-aggregate:row-retrieve (netlist-aggregate 'row:retrieve))
(define netlist-non-aggregate
  ((*board* 'open-table) 'netlist-non-aggregate #t))
(define netlist-non-aggregate:row-retrieve
  (netlist-non-aggregate 'row:retrieve))
(define resistor-nets ((*board* 'open-table) 'resistor-nets #t))
(define resistor-nets:row-retrieve (resistor-nets 'row:retrieve))

(define (db:read-net name)
  (require 'scanf)
  (call-with-input-file (if (has-suffix? name ".net")
			    name
			    (string-append name ".net"))
    (lambda (port)
      (define (read1)
	(let ((line (read-line port)))
	  (cond ((eof-object? line) #f)
		((string=? line "$END") #f)
		((zero? (string-length line)) #f)
		((char=? #\, (string-ref line (- (string-length line) 1)))
		 (string-set! line (- (string-length line) 1) #\ )
		 (cons line (read1)))
		(else (list line)))))
      ;; Returns list of (NAME INDEX SHEET)
      ;; NAME a string, INDEX and SHEET either #f or integers.
      (define (parse-name str)
	(let* ((n (string-length str))
	       (i/ (string-reverse-index str #\/))
	       (sheet (and i/ (string->number (substring str (+ i/ 1) n))))
	       (iend (or i/ n)))
	  (if sheet			;otherwise unnamed net
	      (let loop ((i iend))
		(cond ((zero? i)
		       (list (substring str 0 iend) #f sheet))
		      ((char-numeric? (string-ref str (- i 1)))
		       (loop (- i 1)))
		      ((> iend i)
		       (list (substring str 0 i)
			     (string->number (substring str i iend))
			     sheet))
		      (else
		       (list (substring str 0 i) #f sheet))))
	      (list str #f #f))))
      ;; Returns list (SIGNAL-NAME ((PTAG . PIN) ...))
      ;; where SIGNAL-NAME and PTAG are symbols, PIN an integer.
      (define (read-net-entry)
	(let ((s (read1)))
	  (and s
	       (call-with-input-string (apply string-append s)
		 (lambda (stp)
		   (let ((net-name #f))
		     (fscanf stp " %[^;]" net-name)
		     (let ((parsed (parse-name net-name)))
		       (or (eq? #\; (read-char stp))
			   (synch:warn "strange netlist lines" s))
		       (let loop ((u-pins '())
				  (r-pins '()))
			 (let ((c (peek-char stp)))
			   (cond
			    ((eof-object? c)
			     (netlist-insert parsed u-pins r-pins)
			     #t)
			    ((char-whitespace? c)
			     (read-char stp)
			     (loop u-pins r-pins))
			    ((char=? c #\U)
			     (let ((u-number #f)
				   (pin #f))
			       (fscanf stp "%[^.].%s" u-number pin)
			       (let ((u-sym (string->symbol
					     (string-downcase u-number)))
				     (pin (and pin
					       (or (string->number pin)
						   (string->symbol
						    (string-downcase pin))))))
				 (let try ((i 0))
				   (cond ((u-ptag:row-retrieve u-sym) =>
					  (lambda (row)
					    (loop
					     (cons (cons (cadr row) pin)
						   u-pins)
					     r-pins)))
					 ((> i 2)
					  (slib:error 'db:read-net
						      "can't register" u-sym))
					 (else
					  (db:register-u-ptag u-sym 'unknown)
					  (try (+ i 1))))))))
			    ((char=? c #\R)
			     (let ((r-number #f))
			       (fscanf stp "R%d.%*d" r-number)
			       (loop u-pins
				     (if r-number
					 (cons r-number r-pins)
					 r-pins))))
			    (else
			     (fscanf stp " %*s")
			     (loop u-pins r-pins))))))))))))
      (define netlist-insert
	(let ((netlist:row-insert (netlist 'row:insert))
	      (netlist-inverse:row-insert (netlist-inverse 'row:insert))
	      (resistor-nets:row-update (resistor-nets 'row:update)))
	  (lambda (parsed u-pins r-pins)
	    (let ((name (string->symbol (string-downcase (car parsed))))
		  (index (cadr parsed))
		  (sheet (caddr parsed)))
	      (netlist:row-insert
	       (list name (or index -1) sheet u-pins r-pins))
	      (netlist-agg-update name index)
	      (for-each (lambda (pin)
			  (netlist-inverse:row-insert
			   (list (car pin) (cdr pin) name index)))
			u-pins)
	      (for-each (lambda (r-num)
			  (let ((row (resistor-nets:row-retrieve r-num)))
			    (resistor-nets:row-update
			     (list r-num
				   (cons name
					 (if row (cadr row) '()))))))
			r-pins)))))
      (define netlist-agg-update
	(let ((agg-update (netlist-aggregate 'row:update))
	      (non-agg-insert (netlist-non-aggregate 'row:insert)))
	  (lambda (name index)
	    (if index
		(let ((agg-row (netlist-aggregate:row-retrieve name)))
		  (agg-update
		   (if agg-row
		       (let ((bounds (cadr agg-row)))
			 (list name
			       (list (min index (car bounds))
				     (max index (cadr bounds)))))
		       (list name (if index (list index index) #f)))))
		(non-agg-insert (list name))))))

      (let search ((line (read-line port)))
	(cond ((eof-object? line)
	       (slib:error 'read-net "end of file" fname))
	      ((string=? line "$NETS")
	       (let loop ()
		 (if (read-net-entry) (loop))))
	      (else (search (read-line port))))))))

(define (db:print-power-connections)
  (define (power? name)
    (or (memq name '(vcc gnd))
	(let* ((str (symbol->string name))
	       (n (string-length str)))
	  (and (> n 2)
	       (memv (string-ref str 0) '(#\+ #\-))
	       (char=? (string-ref str (- n 1)) #\v)))))
  ;; Returns list of power or ground net names.
  (define (net-connects row)
    (apply
     (lambda (name index sheet u-pins r-pins)
       (let loop ((r-nets (apply append
				 (map (lambda (r)
					(cadr (resistor-nets:row-retrieve r)))
				      r-pins)))
		  (p-nets '()))
	 (cond ((null? r-nets)
		p-nets)
	       ((memq (car r-nets) p-nets)
		(loop (cdr r-nets)
		      p-nets))
	       ((power? (car r-nets))
		(loop (cdr r-nets)
		      (cons (car r-nets) p-nets)))
	       (else
		(loop (cdr r-nets)
		      p-nets)))))
     row))
  (define (print-net name bounds)
    (let ((connects '()))
      ((netlist 'for-each-row)
       (lambda (row)
	 (set! connects (union (net-connects row) connects)))
       name #f)
      (printf "%10s:"
	      (cond ((not bounds)
		     name)
		    ((= (car bounds) (cadr bounds))
		     (symbol-append name (car bounds)))
		    (else
		     (sprintf #f "%s[%d:%d]" name 
			      (car bounds) (cadr bounds)))))
      (for-each
       (lambda (x) (printf " %7s" x))
       connects))
    (newline))

  ((netlist-aggregate 'for-each-row)
   (lambda (row)
     (apply print-net row)))

  ((netlist-non-aggregate 'for-each-row)
   (lambda (row)
     (print-net (car row) #f))))


(define (merge-bounds lst alst)
  (let ((pair (assq (car lst) alst)))
    (cond ((not pair) (cons lst alst))
	  ((null? (cdr pair))
	   (report 'multiple-declaration-of pair)
	   alst)
	  ((null? (cdr lst))
	   (report 'multiple-declaration-of lst)
	   alst)
	  ((not (cadr pair))
	   (report 'replicated 'array? lst))
	  ((< (cadr pair) (caddr pair))
	   (set-car! (list-tail pair 1) (min (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (max (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (append (cadddr pair) (cadddr lst)))
	   alst)
	  (else
	   (set-car! (list-tail pair 1) (max (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (min (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (append (cadddr lst) (cadddr pair)))
	   alst))))

(define (get-bounds name)
  (let ((vb (vector-bounds name)))
    (if vb (append vb (list (list name))) (list name))))

(define (call-with-classified-pins tag block proc)
  (define outputs '())
  (define inouts '())
  (define inputs '())
  (define registers '())
  (define *lift-bus* (provided? 'lift-bus))
  ((pin-table 'for-each-row)
   (lambda (pin-row)
     (apply
      (lambda (ptag name pin-type pin-number output-enable paired-input)
	(let ((lst (get-bounds name)))
	  (cond
	   ((or (not block) (signal-table:row-retrieve block name))
	    (case pin-type
	      ((output)
	       (set! outputs (merge-bounds lst outputs))
	       (set! registers (merge-bounds lst registers)))
	      ((input)
	       (set! inputs (merge-bounds lst inputs)))
	      ((input/output)
	       (cond ((and (eq? paired-input name)
			   (not (eq? #t output-enable)))
		      (set! paired-input (symbol-append paired-input '- 'pin))))
	       (set! registers (merge-bounds lst registers))
	       (set! inouts
		     (merge-bounds (get-bounds paired-input) inouts))
	       (cond
		(*lift-bus*
		 (if (symbol? output-enable)
		     (set! outputs
			   (merge-bounds (get-bounds output-enable) outputs))
		     (print 'output-enable 'not 'symbol? output-enable))
		 (set! inputs (merge-bounds (get-bounds paired-input) inputs))
		 (set! outputs (merge-bounds lst outputs)))))
	      ((internal)
	       (let* ((shape (array-bounds block name))
		      (newb (if (pair? shape)
				(if (null? (cdr lst))
				    (append lst '(#f #f) shape)
				    (append lst shape))
				lst)))
		 (cond (block
			(if (used-by-other-blocks? name block)
			    (set! outputs (merge-bounds newb outputs)))
			(set! registers (merge-bounds newb registers))))))
	      ((wire-and wire-or)
	       (report "found wire" pin-row))))
	   ((used-in-table:get-block name block)
	    (set! inputs (merge-bounds lst inputs))))))
      pin-row))
   tag)
  (proc outputs inouts inputs registers))

;; Given a list of vectorized signal names SIGS, return an expression
;; combining all into one vectorized signal.
(define (bus-expression sigs)
  (let* ((bnds (sort (map get-bounds sigs)
		     (lambda (b1 b2)
		       (< (cadr b1) (cadr b2)))))
	 (bnd1 (car bnds)))
    (define (mult a b)
      (cond ((eqv? 1 a) b)
	    ((eqv? 1 b) a)
	    (else `(* ,a ,b))))
    (define (add a b)
      (cond ((eqv? 0 a) b)
	    ((eqv? 0 b) a)
	    (else `(+ ,a ,b))))
    (let recur ((cofs (map (lambda (bnd) (car (cadddr bnd)))
			   bnds))
		(pows (if (< (cadr bnd1) (caddr bnd1))
			  (map cadr bnds)
			  (map caddr bnds)))
		(pow 0))
      (if (null? cofs)
	  0
	  (mult (expt 2 (- (car pows) pow))
		(add (car cofs)
		     (recur (cdr cofs)
			    (cdr pows)
			    (car pows))))))))


;; BOUNDS may be an integer, a 2-list of (LOWER-BOUND UPPER-BOUND), or #F.
;; Output is a list of lists : ((PTAG SIGNAL-NAME) ...)
(define (db:netlist-connects net-name bounds)
  (define (union s1 s2)
    (cond ((null? s1) s2)
	  ((null? s2) s1)
	  ((member (car s1) s2) (union (cdr s1) s2))
	  (else (union (cdr s1) (cons (car s1) s2)))))
  (if (list? bounds)
      (let ((imin (apply min bounds))
	    (imax (apply max bounds)))
	(let loop ((i imin)
		   (ret '()))
	  (if (> i imax)
	      ret
	      (loop (+ i 1)
		    (union (db:netlist-connects net-name i) ret)))))
      (let ((row (netlist:row-retrieve net-name (or bounds -1))))
	(and row
	     (apply
	      (lambda (name index sheet u-pins r-pins)
		(let loop ((us u-pins)
			   (ret '()))
		  (if (null? us)
		      ret
		      (loop
		       (cdr us)
		       (let ((pindex-row (pindex:row-retrieve (caar us) (cdar us))))
			 (if pindex-row
			     (cons (list (caar us) (list-ref pindex-row 4)) ret)
			     ret))))))
	      row)))))

;   ================================================================

;;(trace synch:define-signal synch:define-pin synch:register-block)
;;(ta "simsync.scm")

;; Output binary format.

(define (dump:spew . args)
  (define len (length args))
  (display #\@ pf:dumpport)
  (display (number->string dump:address 16) pf:dumpport)
  (set! dump:address (+ dump:address len))
  (map (lambda (x)
	 (display #\space pf:dumpport)
	 (display (case x
		    ((#t) #\1)
		    ((#f) #\0)
		    (else (cond ((not (number? x))
				 (report 'dump:spew funny-value x) #\z)
				((negative? x) #\z)
				(else (number->string x 16)))))
		  pf:dumpport))
       args)
  (newline pf:dumpport))
